﻿﻿//------------------------------------------------------------------------------------------------
// BEGIN SELECTORAPI CONSTRUCTOR
//------------------------------------------------------------------------------------------------
function SelectorAPI()
{
    // SelectorAPI enums -------------------------------------------------------------------------
    
    // Token types required to parse normalized selectors
    var TT_Identifier               = 1,    
        TT_AdjacentSelector         = 2,
        TT_Dot                      = 3,
        TT_Colon                    = 4,
        TT_Equal                    = 5,
        TT_Asterisk                 = 6,
        TT_EscColon                 = 7,
        TT_LParen                   = 8,
        TT_RParen                   = 9,
        TT_EOF                      = 10,
        TT_String                   = 11,
        TT_Hash                     = 12,
        TT_Bang                     = 13,
        TT_ChildSelector            = 14,
        TT_WalkAdjacentSelector     = 15,
        TT_AttributeStart           = 16,
        TT_AttributeEnd             = 17,
        TT_SpaceEquals              = 18,
        TT_HyphenEquals             = 19,
        TT_BeginEquals              = 20,
        TT_EndEquals                = 21,
        TT_SubstrEquals             = 22,
        TT_Whitespace               = 23,
        TT_Name                     = 24,
        TT_Unknown                  = 25;
            
    // Symbols to look for
    var CHAR_ESCAPE                 = '\\',
        CHAR_DOT                    = '.',
        CHAR_COLON                  = ':',
        CHAR_LEFT_PAREN             = '(',
        CHAR_RIGHT_PAREN            = ')',
        CHAR_LEFT_ANGLEBRACKET      = '<',
        CHAR_RIGHT_ANGLEBRACKET     = '>',
        CHAR_HASH                   = '#',
        CHAR_ASTERISK               = '*',
        CHAR_EQUAL                  = '=',
        CHAR_HYPHEN                 = '-',
        CHAR_BANG                   = '!',
        CHAR_ATTR_LBRACKET          = '[',
        CHAR_ATTR_RBRACKET          = ']',
        CHAR_ATTR_TILDE             = '~',
        CHAR_ATTR_PIPE              = '|',
        CHAR_PLUS                   = '+',
        CHAR_HAT                    = '^',
        CHAR_DOLLAR                 = '$';

    // Attribute match modes
    var EMatchUnknown               = 0,
        EMatchDefined               = 1,
        EMatchExact                 = 2,
        EMatchList                  = 3,
        EMatchLang                  = 4,
        EMatchBegin                 = 5,
        EMatchEnd                   = 6,
        EMatchSubstr                = 7;
       
    // Combinator/navigation types
    var ENavigationUnknown          = 0,
        ENavigationAncestor         = 1,
        ENavigationChild            = 2,
        ENavigationAdjacent         = 3,
        ENavigationWalkAdjacent     = 4;

    // Pseudoelement types
    var EPelemNone                  = 0,
        EPelemUnknown               = 1,
        EPelemFirstLetter           = 2,
        EPelemFirstLine             = 3;

    // Pseudoclass types
    var EPclassNone                 = 0,
        EPclassUnknown              = 1,
        EPclassLink                 = 2,
        EPclassVisited              = 3,
        EPclassActive               = 4,
        EPclassHover                = 5,
        EPclassFocus                = 6,
        EPclassFirstChild           = 7;


    // SelectorAPI private classes ---------------------------------------------------------------

    //--------------------------------------------------------------------------------------------
    // BEGIN SELECTORTOKENIZER CONSTRUCTOR
    //--------------------------------------------------------------------------------------------
    
    function SelectorTokenizer(strSelector)
    {
        // SelectorTokenizer member variables ----------------------------------------------------        
        
        // These regexp's are per W3C CSS3 selector spec, 
        // with macros (ie "nonascii") already expanded and inlined

        var REGEXP_WHITESPACE = /^[ \t\r\n\f]+/;

        // W3C's "escape" macro slightly modified not to allow "\:" in an identifier
        //                     { ---------------------------- nmstart ------------------------------------ }{ ------------------------------ nmchar --------------------------------------- }
        //                     { ltrs |  nonascii  |             unicode                |     escape       }{ ltr/nums |  nonascii  |             unicode                |     escape       }
        var REGEXP_IDENT = /^-?([_a-z]|[^\x00-\x7F]|(\\[0-9a-f]{1,6}(\r\n|[ \n\r\t\f])?)|\\[^\n\r\f0-9a-f:])([_a-z0-9-]|[^\x00-\x7F]|(\\[0-9a-f]{1,6}(\r\n|[ \n\r\t\f])?)|\\[^\n\r\f0-9a-f:])*/i;

        //                  { ------------------------------ nmchar --------------------------------------- }
        //                  { ltr/nums |  nonascii  |             unicode                |     escape       }
        var REGEXP_NAME = /^([_a-z0-9-]|[^\x00-\x7F]|(\\[0-9a-f]{1,6}(\r\n|[ \n\r\t\f])?)|\\[^\n\r\f0-9a-f:])+/i;

        //                      Q(  letters    | escpd newlines  |  nonascii  |             unicode                |     escape      )  Q
        var REGEXP_STRING1 = /^\"([^\n\r\f\\\"]|\\(\n|\r\n|\r|\f)|[^\x00-\x7F]|(\\[0-9a-f]{1,6}(\r\n|[ \n\r\t\f])?)|\\[^\n\r\f0-9a-f])*\"/i;
        var REGEXP_STRING2 = /^\'([^\n\r\f\\\']|\\(\n|\r\n|\r|\f)|[^\x00-\x7F]|(\\[0-9a-f]{1,6}(\r\n|[ \n\r\t\f])?)|\\[^\n\r\f0-9a-f])*\'/i;

        // SelectorTokenizer member functions ----------------------------------------------------

        // Scan for the next recognized token and return an enum type 
        // representing what we found. If the token is a text type (ident or string)
        // we will hold onto the text and the user will be able to access it 
        // through GetLastText()
        function NextToken()
        {
            var ttRet = TT_Unknown;
            var strRemaining = this._strSelector.substring(this._offset);
            var strNextChar, strNextTwoChars;
                
            if(this._offset >= this._strSelector.length)
            {
                ttRet = TT_EOF;
            }
            else if(strRemaining.match(REGEXP_WHITESPACE))
            {
                ttRet = TT_Whitespace;
                this._offset += RegExp.lastMatch.length;
            }
            else if(strRemaining.match(REGEXP_IDENT))
            {
                ttRet = TT_Identifier;
                this._strLastText = RegExp.lastMatch;
                this._offset += RegExp.lastMatch.length;
            }
            else if(strRemaining.match(REGEXP_NAME))
            {
                ttRet = TT_Name;
                this._strLastText = RegExp.lastMatch;
                this._offset += RegExp.lastMatch.length;
            }
            else if(strRemaining.match(REGEXP_STRING1) ||
                    strRemaining.match(REGEXP_STRING2))
            {
                ttRet = TT_String;
                var str = RegExp.lastMatch;
	            this._strLastText = str.substr(1, str.length - 2); // Pluck off quotes
                this._offset += RegExp.lastMatch.length;
            } 
            else
            {
                // At this point we're just looking for single chars or two-char pairs                
                strNextChar = strRemaining.substr(0,1);
                // It's ok if there's no char after the next. Tests will just fail (as they should)
                strSecondChar = strRemaining.substr(1,1);
                
                switch(strNextChar)
                {
                    case CHAR_DOT:
                        ttRet = TT_Dot;
                        break;                        
                    case CHAR_COLON:
                        ttRet = TT_Colon;
                        break;                        
                    case CHAR_ESCAPE:
                        if(CHAR_COLON == strSecondChar)
                        {
                            ttRet = TT_EscColon;
                            this._offset++; // since we matched two chars
                        }
                        else
                        {
                            // DebugPrint("dont know what to do with backslash");
                            ttRet = TT_Unknown;
                        }
                        break;   
                    case CHAR_LEFT_PAREN:
                        ttRet = TT_LParen;
                        break;
                    case CHAR_RIGHT_PAREN:
                        ttRet = TT_RParen;
                        break;
                    case CHAR_LEFT_ANGLEBRACKET:
                        // DebugPrint("Found < but I dont know what to do with it");
                        ttRet = TT_Unknown;
                        break;
                    case CHAR_RIGHT_ANGLEBRACKET:
                        ttRet = TT_ChildSelector;
                        break;
                    case CHAR_HASH:
                        ttRet = TT_Hash;
                        break;         
                    case CHAR_ASTERISK:
                        if(CHAR_EQUAL == strSecondChar)
                        {
                            ttRet = TT_SubstrEquals;
                            this._offset++; // since we matched two chars
                        }
                        else
                        {
                            ttRet = TT_Asterisk;
                        }
                        break;         
                    case CHAR_EQUAL:
                        ttRet = TT_Equal;
                        break;         
                    case CHAR_HYPHEN:
                        // DebugPrint("Don't know what to do with hyphen");
                        ttRet = TT_Unknown; 
                        break;         
                    case CHAR_BANG:
                        ttRet = TT_Bang;
                        break;         
                    case CHAR_ATTR_LBRACKET:
                        ttRet = TT_AttributeStart;
                        break;         
                    case CHAR_ATTR_RBRACKET:
                        ttRet = TT_AttributeEnd;
                        break;         
                    case CHAR_ATTR_TILDE:
                        if(CHAR_EQUAL == strSecondChar)
                        {
                            ttRet = TT_SpaceEquals;
                            this._offset++; // since we matched two chars
                        }
                        else
                        {
                            ttRet = TT_WalkAdjacentSelector;
                        }
                        break;       
                    case CHAR_ATTR_PIPE:
                        if(CHAR_EQUAL == strSecondChar)
                        {
                            ttRet = TT_HyphenEquals;
                            this._offset++; // since we matched two chars
                        }
                        else
                        {
                            // DebugPrint("UHOH, I don't know what to do with | by itself");
                            ttRet = TT_Unknown;
                        }
                        break;         
                    case CHAR_PLUS:
                        ttRet = TT_AdjacentSelector;
                        break;         
                    case CHAR_HAT:
                        if(CHAR_EQUAL == strSecondChar)
                        {
                            ttRet = TT_BeginEquals;
                            this._offset++; // since we matched two chars
                        }
                        else
                        {
                            // DebugPrint("UHOH, I don't know what to do with ^ by itself");
                            ttRet = TT_Unknown;
                        }
                        break;
                    case CHAR_DOLLAR:
                        if(CHAR_EQUAL == strSecondChar)
                        {
                            ttRet = TT_EndEquals;
                            this._offset++; // since we matched two chars
                        }
                        else
                        {
                            // DebugPrint("UHOH, I don't know what to do with $ by itself");
                            ttRet = TT_Unknown;
                        }               
                        break;
                    default:
                        // DebugPrint("UHOH! Can't idenfity this token: " + strNextChar);
                        ttRet = TT_Unknown;
                        break;
                        
                }
                this._offset++;
            }
            
            this._currToken = ttRet;
            return ttRet;
        }

        // Peek at the next token without advancing to it
        function PeekNextToken()
        {
            // Store current state, call NextToken(), and "undo" the token advance    
            var savedOffset = this._offset;
            var savedLastText = this._strLastText;
            var ttRet = this.NextToken();
            this._offset = savedOffset;
            this._strLastText = savedLastText;
            return ttRet;
        }

        // Return the last text we saw. 
        // Used when NextToken() returns TT_String or TT_Ident
        function GetLastText()
        {
            return this._strLastText;
        }

        function GetCurrToken()
        {
            return this._currToken;
        }

        function NextNonWSToken()
        {
            var tt = this.NextToken();
            while(TT_Whitespace == tt)
            {
                tt = this.NextToken();
            }    
            return tt;
        }     
          
        // SelectorTokenizer member assignment ---------------------------------------------------
    
        this._strSelector = strSelector;
        
        this._offset = 0;
        this._currToken = TT_Unknown;
        this._strLastText = "";
            
        this.NextToken = NextToken;
        this.PeekNextToken = PeekNextToken;
        this.NextNonWSToken = NextNonWSToken;
        this.GetLastText = GetLastText;
        this.GetCurrToken = GetCurrToken;    
    }

    //--------------------------------------------------------------------------------------------
    // END SELECTORTOKENIZER CONSTRUCTOR
    //--------------------------------------------------------------------------------------------

    //--------------------------------------------------------------------------------------------
    // BEGIN SELECTORPARSER CONSTRUCTOR
    //--------------------------------------------------------------------------------------------

    function SelectorParser()
    {
        // SelectorParser member functions -------------------------------------------------------
        
        // Parse an array of rule wrappers and return them as
        // an array of selector objects
        function Parse(aRules)
        {
            var i;
            var collSelectors = selectorFactory.createSelectorCollection();
                
            for(i = 0; i < aRules.length; i++)
            {
                var curRule = aRules[i];
                
                // Filter out selectors with the wrong media
                var media = aRules[i].parentStyleSheet.media;
                if(null != media) media = media.toLowerCase();

                if(null == media || "" == media || -1 != media.indexOf("screen") || "all" == media)
                {
                    var sel = this.ParseSingleSelector(curRule);
                    collSelectors.Add(sel);
                }
                    
            }    
            
            return collSelectors;
             
        } // end Parse


        // Parse a single rule wrapper and return it as a selector object
        function ParseSingleSelector(ruleWrapper)
        {
            var strSelectorText = ruleWrapper.rule.selectorText;

            var selCurrent = this.ParseString(strSelectorText);
            
            // By setting the userData field to our rule wrapper,
            // we will be able to trace back the selector this rule
            // and also know the origin of the rule
            selCurrent.userData = ruleWrapper;
            
            return selCurrent;

        } // end ParseSingleSelector


        // Parse a string of text and return it as a selector object
        function ParseString(strSelectorText)
        {
            var tok = new SelectorTokenizer(strSelectorText);
            
            var strLastText, selNewSelector, tt, ttNext;
            var fStringy, eMatchType, eCombinatorType;
            var selCurrent, selNext, selNewAttr;
            var ePseudoclassType, selNewPclass;
            
            selCurrent = selectorFactory.createSelector();
            
            tt = tok.NextToken();
            
            while(TT_EOF != tt) // This is the "simple selector" loop
            {
                // First parse the element type
                if(TT_Identifier == tt  || TT_Asterisk == tt )
                {
                    strLastText = (TT_Asterisk == tt) ? "*" : tok.GetLastText();
                    
                    tt = tok.NextToken();
                    
                    if(TT_EscColon == tt)
                    {
                        // The ident we found before was the namespace.
                        selCurrent.nmsp = strLastText;
                        
                        // Now grab the actual element type
                        tt = tok.NextToken();
                        if(TT_Identifier == tt || TT_Asterisk == tt)
                        {
                           strLastText = (TT_Asterisk == tt) ? "*" : tok.GetLastText();
                           
                           selCurrent.tagName = strLastText;
                           tt = tok.NextToken();
                        }
                    }
                    else
                    {
                        // No namespace. The ident we found before is the elem type
                        selCurrent.tagName = strLastText;                
                    }        
                }
                
                // Now parse the rest: id's, attributes, classes, pseudos
                while(TT_Hash == tt || TT_Dot == tt || TT_AttributeStart == tt || TT_Colon == tt)
                {
                    if(TT_Hash == tt) // ID
                    {
                        tt = tok.NextToken();
                        if(TT_Identifier == tt || TT_Name == tt)
                        {
                            strLastText = tok.GetLastText();                    
                            selCurrent.id = strLastText;
                        }
                    }
                    else if(TT_Dot == tt) // Class
                    {
                        tt = tok.NextToken();
                        if(TT_Identifier == tt)
                        {
                            strLastText = tok.GetLastText();                    
                            selCurrent.AddClass(strLastText);
                        }
                    }
                    else if(TT_AttributeStart == tt) // Attrib
                    {
                        // Gobble whitespace immediately after bracket
                        tt = tok.NextNonWSToken();
                        
                        // Don't make same mistake IE made :)
                        fStringy = false;
                        
                        if(TT_Identifier == tt )
                        {
                            strLastText = tok.GetLastText();
                            
                            selNewAttr = selectorFactory.createAttributeSelector();
                            selNewAttr.name = strLastText;
                            selCurrent.AddAttribute(selNewAttr);
                            
                            // Gobble whitespace immediately after attr name
                            tt = tok.NextNonWSToken();
                            
                            // Take care of the attribute match type
                            switch(tt)
                            {
                                case TT_BeginEquals:
                                    eMatchType = EMatchBegin;
                                    break;
                                case TT_EndEquals:
                                    eMatchType = EMatchEnd;
                                    break;
                                case TT_SubstrEquals:
                                    eMatchType = EMatchSubstr;
                                    break;
                                case TT_Equal:
                                    eMatchType = EMatchExact;
                                    break;
                                case TT_SpaceEquals:
                                    eMatchType = EMatchList;
                                    break;
                                case TT_HyphenEquals:
                                    eMatchType = EMatchLang;
                                    break;
                                case TT_AttributeEnd:
                                    // This is just an "existence" attribute.
                                    eMatchType = EMatchDefined;
                                    break;                            
                                default:
                                    // DebugPrint("uhoh! dont recognize this match type");
                                    eMatchType = EMatchUnknown;
                                    break;                                                
                            }
                            
                            selNewAttr.matchType = eMatchType;
                            
                            if(TT_AttributeEnd != tt)
                            {
                                // Have a name and operator. Now look for value.
                                tt = tok.NextNonWSToken();
                                
                                if(TT_Identifier == tt || TT_String == tt)
                                {
                                    strLastText = tok.GetLastText();                                    
                                    fStringy = (TT_String == tt);                                    
                                    selNewAttr.value = strLastText;
                                }                                
                                tt = tok.NextNonWSToken();
                                // tt should now point to the closing bracket
                            }                   
                                                
                        } // end if TT_Identifier == tt                
                    }
                    else if(TT_Colon == tt) // Pseudo
                    {
                        tt = tok.NextToken();
                        
                        if(TT_Identifier == tt)
                        {
                            strLastText = tok.GetLastText();                            
                            ePseudoclassType = EPclassNone;                                        

                            switch(strLastText.toLowerCase())
                            {
                                // Pseudoclass and pseudoelement cases are mixed together
                                // This is kinda messy, but we can't tell them apart
                                
                                case "first-line":                            
                                    selCurrent.pseudoElement = EPelemFirstLine;
                                    break;
                                case "first-letter":
                                    selCurrent.pseudoElement = EPelemFirstLetter;
                                    break;
                                case "active":
                                    ePseudoclassType = EPclassActive;
                                    break;
                                case "visited":
                                    ePseudoclassType = EPclassVisited;
                                    break;
                                case "link":
                                    ePseudoclassType = EPclassLink;
                                    break;
                                case "hover":
                                    ePseudoclassType = EPclassHover;
                                    break;
                                case "focus":
                                    ePseudoclassType = EPclassFocus;
                                    break;                  
			                    case "first-child":
				                 ePseudoclassType = EPclassFirstChild;
			                        break;
                                default:
                                    ePseudoclassType = EPclassUnknown;
                                    break;
			                          
                            }             
                            
                            // Attach a pseudoclass selector if we need one
                            if(EPclassNone != ePseudoclassType)
                            {
                                selNextPclass = selectorFactory.createPseudoclassSelector();
                                selNextPclass.pseudoclassType = ePseudoclassType;
                                selCurrent.addPseudoclass(selNextPclass);
                            }       
                        }
                        else if(TT_Colon == tt)
                        {
                            // CSS3-style pseudo-elements (ie "::first-line")
                            tt = tok.NextToken();
                            if(TT_Identifier == tt)
                            {
                                strLastText = tok.GetLastText();

                                switch(strLastText.toLowerCase())
                                {
                                    case "first-line":                            
                                        selCurrent.pseudoElement = EPelemFirstLine;
                                        break;
                                    case "first-letter":
                                        selCurrent.pseudoElement = EPelemFirstLetter;
                                        break;
                                }                        
                            }
                        }            
                    } // end else if TT_Colon == tt
                    
                    tt = tok.NextToken();     
                    
                } // end while class/hash/attr/pseudo
                
                // Just completed parsing a simple selector.
                // Now look for a combinator.
                eCombinatorType = ENavigationUnknown;
                
                // Eat whitespace before combinator, but only if
                // the thing after the ws really is a combinator (or EOF)
                if(TT_Whitespace == tt)
                {
                    ttNext = tok.PeekNextToken();
                    if(TT_ChildSelector == ttNext || 
                       TT_AdjacentSelector == ttNext || 
                       TT_WalkAdjacentSelector == ttNext ||
                       TT_EOF == ttNext)
                       {
                            tt = tok.NextToken();
                       }
                }
                
                switch(tt)
                {
                    case TT_Whitespace:
                        eCombinatorType = ENavigationAncestor;
                        break;                        
                    case TT_ChildSelector:
                        eCombinatorType = ENavigationChild;
                        break;                        
                    case TT_AdjacentSelector:
                        eCombinatorType = ENavigationAdjacent;
                        break;                        
                    case TT_WalkAdjacentSelector:
                        eCombinatorType = ENavigationWalkAdjacent;
                        break;                        
                    case TT_EOF:
                        break;                        
                    default:
                        // DebugPrint("I don't know what to do with this token: " + tt);
                        break;
                }
                
                tt = tok.NextNonWSToken();
                if(TT_EOF == tt)
                {
                    // DebugPrint("~ EOF");
                }    
                else
                {
                    // Prepare for the next selector in the sequence     
                    
                    selNext = selectorFactory.createSelector();
                    selNext.parent = selCurrent;
                    selNext.navType = eCombinatorType;
                    selNext.userData = selCurrent.userData;
                    selCurrent = selNext;                      
                }
            
            } // end while not EOF
            
            return selCurrent;    

        } // end ParseString
    
    
        // SelectorParser member assignment ------------------------------------------------------
    
        this.Parse = Parse;
        this.ParseSingleSelector = ParseSingleSelector;
        this.ParseString = ParseString;
    }

    //--------------------------------------------------------------------------------------------
    // END SELECTORPARSER CONSTRUCTOR
    //--------------------------------------------------------------------------------------------
    

	// SelectorAPI member variables --------------------------------------------------------------

	var parser, collSelectors;
	
	// SelectorAPI member functions --------------------------------------------------------------


    function ParseDoc(doc)
    {
        var aRules = GatherRulesFromDoc(doc);
        collSelectors = parser.Parse(aRules);
        
        // Set quirks mode if necessary
        if("BackCompat" == doc.compatMode)
        {
            for(var i = 0; i < collSelectors.Count; i++)
            {
                collSelectors.item(i).SetQuirksMode(true);
            }
        }
        
    }

    // Constructor for a RuleWrapper object
    // Stores a rule object plus some additional origin
    // info so we can trace back where this rule came from
    function RuleWrapper(rule, parentStyleSheet, index) 
    {
        // Rule object provided by object model
        this.rule = rule;
        
        // Pointer to styleSheet object that provided this rule
        this.parentStyleSheet = parentStyleSheet;
        
        // Index within that sheet of this rule
        this.index = index;
        
    }

    // Return an array with all the rules from all the stylesheets in the doc
    function GatherRulesFromDoc(doc)
    {
        var aRules = new Array();
        var i, ssCurrentSheet;
            
        if(!doc.styleSheets) return aRules;	

        for(i = 0; i < doc.styleSheets.length; i++)
        {
            ssCurrentSheet = doc.styleSheets(i);
            ssCurrentSheet.index = i;
            AppendRules(ssCurrentSheet, aRules, 0);        
        }   
           
        return aRules;    
    }

    // Append to aRules all the rules in the given style sheet
    function AppendRules(ssCurrentSheet, aRules, nRecursionDepth)
    {
        var i, ssCurrentImport, rulCurrentRule;

        // Mostly a sanity check
        if(nRecursionDepth > 10)
        {
            return;
        }
        
        // Order matters. Add imported sheets before local rules.
        for(i = 0; i < ssCurrentSheet.imports.length; i++)
	    {
	        ssCurrentImport = ssCurrentSheet.imports(i);
	        ssCurrentImport.index = i;
		    AppendRules(ssCurrentImport, aRules, nRecursionDepth + 1);
	    }
    	
	    for(i = 0; i < ssCurrentSheet.rules.length; i++)
	    {
	        rulCurrentRule = new RuleWrapper(ssCurrentSheet.rules(i), ssCurrentSheet, i);
	        aRules.push(rulCurrentRule);
	    }    		
    }

    function GetNodeString(currNode)
    {
        var strRet = currNode.nodeName;
        if(currNode.className.length > 0) strRet += " class=" + currNode.className;
        if(currNode.id.length > 0) strRet += " id=" + currNode.id;
        return strRet;
    }
    
    function TraceAll(node)
    {
        var i, j;
        var aRet = new Array();
        var aSortMe;
        
        //
        // Match all selectors against this node and its ancestors
        //

        var currNode = node;
        
        retTitle = "Styles applied to " + GetNodeString(node);

        // Go up the chain of ancestors, BUT...
        // there's a node above the html node. We want to avoid it.
        while(null != currNode && null != currNode.parentNode)
        {
            // Get the selectors that match the current node
            var collMatches = engine.getMatchingRules(currNode, collSelectors);
                
            if(currNode.style.cssText.length > 0 || collMatches.Count > 0)
            {
                 aRet.push("/****************************************************************************/");
                 aRet.push("/* Matches for node: " + GetNodeString(currNode) + " */");
                 aRet.push("/****************************************************************************/\r\n");
                 
            }

            // Show inline rule if we have one   
            if(currNode.style.cssText.length > 0)
            {
                var selTemp = selectorFactory.createSelector();
                selTemp.userData = new RuleWrapper(currNode, null, -1);
                aRet.push(OutputSelector(selTemp));
            }

            aSortMe = new Array();
            for(j = 0; j < collMatches.Count; j++)
            {
                aSortMe.push(collMatches.item(j));
            }

            aSortMe.sort(SpecificitySortFunction);
            
            for(j = 0; j < aSortMe.length; j++)
            {
                aRet.push(OutputSelector(aSortMe[j]));
            }
            
            currNode = currNode.parentNode;            
        }

        

        // Store text and title for later match
        if(0 == aRet.length)
        {
            retCssText = "No match."            
        }
        else
        {
            retCssText = aRet.join("\r\n");
        }

        return true;
    }

    function SpecificitySortFunction(a,b)
    {
        return b.GetSpecificity() - a.GetSpecificity();
    }

    function GetHeader(sel)
    {
        var strRet = "/* ";
        
        var idx = sel.userData.index;
        var ss = sel.userData.parentStyleSheet;
        if(null == ss)
        {
            strRet += "Inline style";
        }
        else
        {
            var ssStr = (ss.href) ? ss.href : "embedded stylesheet";
            strRet += "Rule " + (idx+1) + " of " + ssStr;
        }

	    strRet += " */ \r\n";

        return strRet;        
    }


    function OutputSelector(currSel)
    {
        var strOutput = "";
        var getstring = currSel.GetString();
        
        strOutput += GetHeader(currSel);
        strOutput += getstring;
        if(getstring.length > 0) strOutput += " ";
        strOutput += "{\r\n\t" + FormatRuleBody(currSel.userData.rule.style.cssText) + "\r\n}\r\n";

        return strOutput;	
    }

    // Look for a property within a cssText block
    function GetIndexOfPropHelper(cssText, prop)
    {
        // return cssText.indexOf(prop + ":");
    
        var prop2 = prop + ":";
    
        // if we are given prop = "width" then look for "width:" in the cssText
        // but don't produce a false positive for things like "min-width:" 
        
        if(null != cssText.match(eval("/^" + prop2 + "/")) ||
           null != cssText.match(eval("/[^-\w]" + prop2 + "/")))
        {
            return RegExp.index;
        }
        else return -1;
    }
    
    function GetIndexOfProp(cssText, prop)
    {
        // Handle shorthand properties in this function 
        // (border is shorthand for border-top-width, etc)
                
        // First try the full name
        var helperResult = GetIndexOfPropHelper(cssText, prop);
        
        if(-1 != helperResult)
        {
            return helperResult;
        }
        
        // If that failed, handle special cases
        if(-1 != prop.indexOf("-COLOR"))
        {
            helperResult = GetIndexOfPropHelper(cssText, "COLOR");
            if(-1 != helperResult)
            {
                return helperResult;
            }            
        }
        if("LINE-HEIGHT" == prop)
        {
            helperResult = GetIndexOfPropHelper(cssText, "FONT");
            if(-1 != helperResult)
            {
                return helperResult;
            }
        }
        
        // If no special cases, try plucking off hyphens        
        // So border-top-width becomes border-top and then border
        var idx;
        
        while(-1 != (idx = prop.indexOf("-")))
        {
            prop = prop.substr(0, idx);
            helperResult = GetIndexOfPropHelper(cssText, prop);
            if(-1 != helperResult)
            {
                return helperResult;
            }
        }
        
        // No luck
        return -1;    
                
    }

    function IsPropertyInStyle(style, prop)
    {
        if(null == prop || "" == prop)
        {
            return true;
        }
        else        
        {
            return ( -1 != GetIndexOfProp(style.cssText, prop));	
        }
    }
    
    function IsImportant(cssText, prop)
    {
        var ret = false;
        var idxImp, idxSemi, idxProp = 0;
        
        // Find where this property begins
        idxProp = GetIndexOfProp(cssText, prop);
        
        // Then, starting from that property, see if a
        // !important declaration appears before the next semicolon
        
        idxImp = cssText.indexOf("! important", idxProp);
        idxSemi = cssText.indexOf(";", idxProp);
        
        if(-1 == idxImp)
        {
            ret = false;
        }
        else if(-1 == idxSemi)
        {
            ret = true;
        }
        else
        {
            ret = (idxImp < idxSemi)
        }
            
        return ret;
        
    }
    
    
    
    
    function TraceProperty(node, prop)
    {
        var i;
        var collPotentialMatches = selectorFactory.createSelectorCollection();
    
        prop = prop.toUpperCase();
        
        
                
        //	
        // Filter out selectors that don't have the desired property
        //
        for(i = 0; i < collSelectors.Count; i++)
        {
            var currSelector = collSelectors(i);

            if(IsPropertyInStyle(currSelector.userData.rule.style, prop))
            {
                collPotentialMatches.Add(currSelector);
            }
        }
        
        //
        // Match all selectors against this node and its ancestors
        //

        var currNode = node, matchedNode = null;

        var maxSel = null;  // Selector of maximal specificity found so far
        var maxSpec = 0;    // Maximal specificity found so far
        var maxImp = 0;     // Maximal importance (0 or 1) found so far
        
        
        // Go up the chain of ancestors, BUT...
        // there's a node above the html node. We want to avoid it.
        while(null != currNode && null != currNode.parentNode && null == maxSel)
        {
            // If we have an inline rule, it takes precendence. Don't bother running the match.            
            if(IsPropertyInStyle(currNode.style, prop))
            {
                maxSel = selectorFactory.createSelector();
                maxSel.userData = new RuleWrapper(currNode, null, -1); 
                matchedNode = currNode;                               
            }
            else
            {
                // Our starting point will be the selectors that match the current node
                var collMatches = engine.getMatchingRules(currNode, collPotentialMatches);

                maxImp = 0;               

                for(var j = 0; j < collMatches.Count; j++)
                {
                    var currSel = collMatches.item(j);
                
                    var currImp = IsImportant(currSel.userData.rule.style.cssText, prop) ? 1 : 0;

                    if(currImp > maxImp || 
                      (currImp == maxImp && currSel.GetSpecificity() >= maxSpec))
                    {
                        maxSel = currSel;
                        maxSpec = currSel.GetSpecificity();
                        maxImp = currImp;
                        matchedNode = currNode;
                    }
                }

                currNode = currNode.parentNode;
            }
        }

        //
        // Store the match info for later retrieval by dev toolbar
        //
        if(null == maxSel)
        {
            retCssText = "No match."
            retTitle = "Style tracer";
            retLineNum = 0;
        }
        else
        {
            StoreMatchInfo(maxSel.userData);
        }
        
        return true;
    }

    function FormatRuleBody(cssText)
    {
        var ret = cssText.replace(/\!.important/g, " !important");
        ret = ret.replace(/; /g, ";\r\n\t");
        return ret;
    }
        
    function StoreMatchInfo(ruleWrapper)
    {
        // We are given the rulewrapper for the responsible style code for the last trace.
        //
        // Stash away the following:
        // 1. cssText of the file containing the responsible rule (formatted prettily)
        // 2. title to use for the window (most commonly the URL of the stylesheet, or possibly a message like "(embedded style)"
        // 3. line number of the responsible rule within that file
        // 4. character number of the responsible rule (used only within the .net app since it can't scroll-into-view on a line)
        //
        // Immediately after running the trace the devtoolbar will grab these three values via accessor functions
        // This gets around the problem of not being able to return > 1 value, without writing a special container object
    
        if(null == ruleWrapper.parentStyleSheet)
        {
            retTitle = "(Inline style)";            
            
            retCssText = "{\r\n\t" + FormatRuleBody(ruleWrapper.rule.style.cssText) + "\r\n}\r\n";
            
            retLineNum = 0;
        }
        else
        {
            var href = ruleWrapper.parentStyleSheet.href;
            retTitle = (null == href || "" == href) ? "(Embedded stylesheet)" : href;

            // Output the url at the top of screen so it can be copied
            retCssText = "/* " + retTitle + " */\r\n\r\n";

            // Prettify the cssText
            retCssText += FormatRuleBody(ruleWrapper.parentStyleSheet.cssText);
            
            // Count opening curly braces to find the line number
            var i = 0;
            var countNow = 0, countWant = ruleWrapper.index + 1;
            retLineNum = 1;    // start on line 1 
            
            while(i < retCssText.length && countNow < countWant)
            {
                switch(retCssText.substr(i,1))
                {
                    case "\r": retLineNum++; break; 
                    case "{": countNow++; break;
                    case "@":  // compensate for the extra "{" of the media block 
                       if("@media" == retCssText.substr(i,6)) countNow--;
                       break;
                }
                
                i++;
            }           
            
            retCharNum = i;

        }
        
    }


    function MatchCountReport(doc)
    {
        var collMatches, i;
        var aRet = new Array();	    

        for(i = 0; i < collSelectors.Count; i++)
        {
            collMatches = engine.matchAll(doc, collSelectors(i));

            aRet.push("/* " + collMatches.Count + " match(es) for: */");
            aRet.push(OutputSelector(collSelectors(i)));
        }

        retTitle = "Match count report";
        
        if(0 == aRet.length)
        {
            retCssText = "No match.";
        }
        else
        {
            retCssText = aRet.join("\r\n");        
        }
        
        return true;

     }

   
    //
    // Add rules that match currNode to aMatches
    //
    function AddMatches(currNode, aMatches)
    {
        var i,j;
        var collMatches = engine.getMatchingRules(currNode, collSelectors);
        
        for(i = 0; i < collMatches.Count; i++)
        {
            // Only insert matches we haven't seen
            
            var fFound = false;
            var currMatch = collMatches.item(i);
            
            for(j = 0; j < aMatches.length; j++)
            {
                if(aMatches[j] == currMatch)
                {
                    fFound = true;
                    break;
                }
            }
            
            if(false == fFound) 
            {
                aMatches.push(currMatch);          
            }
        }  
    }

    //
    // Return a string with the styles for this element, its ancestors, and its descendants
    //
    function GetAllAppliedStyles(node)
    {
        var aMatches = new Array();

        var i;
        
        // Start at the parent since we'll handle the node itself in the descendant portion
        var currNode = node.parentNode;
        
        
        // Go up the chain of ancestors, BUT...
        // there's a node above the html node. We want to avoid it.
        while(null != currNode && null != currNode.parentNode)
        {
            AddMatches(currNode, aMatches);
            currNode = currNode.parentNode;
        }

        // Get styles for descendants
        var aNodeStack = new Array();
        var children;
        
        currNode = node;
        
        while(undefined != currNode)
        {
            AddMatches(currNode, aMatches);
            
            children = currNode.children;
            
            for(i = 0; i < children.length; i++)
            {
                aNodeStack.push(currNode.children(i));
            }
            
            currNode = aNodeStack.pop();
        }
        
        
        // Output the text of all the selectors
        var aRet = new Array();
        
        for(i = 0; i < aMatches.length; i++)
        {
            aRet.push(OutputSelector(aMatches[i]));
        }
        
        if(0 == aRet.length)
        {
            return "/* No matching styles were found. */\r\n";
        }
        else
        {
            return aRet.join("\r\n");
        }
           
    }

    
    // SelectorAPI member assignment -------------------------------------------------------------

    this.ParseDoc = ParseDoc;
    this.TraceAll = TraceAll;
    this.MatchCountReport = MatchCountReport;
    this.TraceProperty = TraceProperty;
    this.GetAllAppliedStyles = GetAllAppliedStyles;

    parser = new SelectorParser();
    this.parser = parser;
   
    



}
//------------------------------------------------------------------------------------------------
// END SELECTORAPI CONSTRUCTOR
//------------------------------------------------------------------------------------------------

// Global methods and variables ------------------------------------------------------------------
var api, elem, doc, prop;
var selectorFactory = null;
var engine = null;

var retCssText, retLineNum, retTitle, retCharNum;

function CreateAPIObject()
{
    api = new SelectorAPI();
}

function ParseDoc()
{
    api.ParseDoc(doc);
}

function MatchCountReport()
{
    return api.MatchCountReport(doc);
}

function TraceProperty()
{
    return api.TraceProperty(elem, prop);
}

function TraceAll()
{
    return api.TraceAll(elem);
}

function GetAllAppliedStyles()
{
    return api.GetAllAppliedStyles(elem);
}

function GetLastMatchCssText()
{
    return retCssText;
}

function GetLastMatchLineNum()
{
    return retLineNum;
}

function GetLastMatchTitle()
{
    return retTitle;
}

// End global methods and variables --------------------------------------------------------------


    
